# Copyright (C) 2005, 2006 by Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

"""Test the uncommit command."""

import os

from bzrlib import uncommit, workingtree
from bzrlib.bzrdir import BzrDirMetaFormat1
from bzrlib.errors import BzrError, BoundBranchOutOfDate
from bzrlib.tests import TestCaseWithTransport


class TestUncommit(TestCaseWithTransport):

    def create_simple_tree(self):
        wt = self.make_branch_and_tree('tree')
        self.build_tree(['tree/a', 'tree/b', 'tree/c'])
        wt.add(['a', 'b', 'c'])
        wt.commit('initial commit', rev_id='a1')

        open('tree/a', 'wb').write('new contents of a\n')
        wt.commit('second commit', rev_id='a2')

        return wt

    def test_uncommit(self):
        """Test uncommit functionality."""
        wt = self.create_simple_tree()

        os.chdir('tree')
        out, err = self.run_bzr('uncommit', '--dry-run', '--force')
        self.assertContainsRe(out, 'Dry-run')
        self.assertNotContainsRe(out, 'initial commit')
        self.assertContainsRe(out, 'second commit')

        # Nothing has changed
        self.assertEqual('a2', wt.last_revision())

        # Uncommit, don't prompt
        out, err = self.run_bzr('uncommit', '--force')
        self.assertNotContainsRe(out, 'initial commit')
        self.assertContainsRe(out, 'second commit')

        # This should look like we are back in revno 1
        self.assertEqual('a1', wt.last_revision())
        out, err = self.run_bzr('status')
        self.assertEquals(out, 'modified:\n  a\n')

    def test_uncommit_checkout(self):
        wt = self.create_simple_tree()

        checkout_tree = wt.bzrdir.sprout('checkout').open_workingtree()
        checkout_tree.branch.bind(wt.branch)

        self.assertEqual('a2', checkout_tree.last_revision())

        os.chdir('checkout')
        out, err = self.run_bzr('uncommit', '--dry-run', '--force')
        self.assertContainsRe(out, 'Dry-run')
        self.assertNotContainsRe(out, 'initial commit')
        self.assertContainsRe(out, 'second commit')

        self.assertEqual('a2', checkout_tree.last_revision())

        out, err = self.run_bzr('uncommit', '--force')
        self.assertNotContainsRe(out, 'initial commit')
        self.assertContainsRe(out, 'second commit')

        # uncommit in a checkout should uncommit the parent branch
        # (but doesn't effect the other working tree)
        self.assertEquals('a1', checkout_tree.last_revision())
        self.assertEquals('a1', wt.branch.last_revision())
        self.assertEquals('a2', wt.last_revision())

    def test_uncommit_bound(self):
        os.mkdir('a')
        a = BzrDirMetaFormat1().initialize('a')
        a.create_repository()
        a.create_branch()
        t = a.create_workingtree()
        t.commit('commit 1')
        t.commit('commit 2')
        t.commit('commit 3')
        b = t.bzrdir.sprout('b').open_branch()
        b.bind(t.branch)
        uncommit.uncommit(b)
        t.set_last_revision(t.branch.last_revision())
        self.assertEqual(len(b.revision_history()), 2)
        self.assertEqual(len(t.branch.revision_history()), 2)
        t.commit('commit 3b')
        self.assertRaises(BoundBranchOutOfDate, uncommit.uncommit, b)
        b.pull(t.branch)
        uncommit.uncommit(b)

    def test_uncommit_revision(self):
        wt = self.create_simple_tree()

        os.chdir('tree')
        out, err = self.run_bzr('uncommit', '-r1', '--force')

        self.assertNotContainsRe(out, 'initial commit')
        self.assertContainsRe(out, 'second commit')
        self.assertEqual('a1', wt.last_revision())
        self.assertEqual('a1', wt.branch.last_revision())

    def test_uncommit_neg_1(self):
        wt = self.create_simple_tree()
        os.chdir('tree')
        out, err = self.run_bzr('uncommit', '-r', '-1', retcode=1)
        self.assertEqual('No revisions to uncommit.\n', out)

    def test_uncommit_merges(self):
        wt = self.create_simple_tree()

        tree2 = wt.bzrdir.sprout('tree2').open_workingtree()

        tree2.commit('unchanged', rev_id='b3')
        tree2.commit('unchanged', rev_id='b4')

        wt.branch.fetch(tree2.branch)
        wt.set_pending_merges(['b4'])
        wt.commit('merge b4', rev_id='a3')

        self.assertEqual('a3', wt.last_revision())
        self.assertEqual([], wt.pending_merges())

        os.chdir('tree')
        out, err = self.run_bzr('uncommit', '--force')

        self.assertEqual('a2', wt.last_revision())
        self.assertEqual(['b4'], wt.pending_merges())

    def test_uncommit_pending_merge(self):
        wt = self.create_simple_tree()
        tree2 = wt.bzrdir.sprout('tree2').open_workingtree()
        tree2.commit('unchanged', rev_id='b3')

        wt.branch.fetch(tree2.branch)
        wt.set_pending_merges(['b3'])

        os.chdir('tree')
        out, err = self.run_bzr('uncommit', '--force')
        self.assertEqual('a1', wt.last_revision())
        self.assertEqual(['b3'], wt.pending_merges())

    def test_uncommit_multiple_merge(self):
        wt = self.create_simple_tree()

        tree2 = wt.bzrdir.sprout('tree2').open_workingtree()

        tree2.commit('unchanged', rev_id='b3')
        wt.branch.fetch(tree2.branch)
        wt.set_pending_merges(['b3'])
        wt.commit('merge b3', rev_id='a3')

        tree2.commit('unchanged', rev_id='b4')
        wt.branch.fetch(tree2.branch)
        wt.set_pending_merges(['b4'])
        wt.commit('merge b4', rev_id='a4')

        self.assertEqual('a4', wt.last_revision())
        self.assertEqual([], wt.pending_merges())

        os.chdir('tree')
        out, err = self.run_bzr('uncommit', '--force', '-r', '2')

        self.assertEqual('a2', wt.last_revision())
        self.assertEqual(['b3', 'b4'], wt.pending_merges())

    def test_uncommit_merge_plus_pending(self):
        wt = self.create_simple_tree()

        tree2 = wt.bzrdir.sprout('tree2').open_workingtree()

        tree2.commit('unchanged', rev_id='b3')
        wt.branch.fetch(tree2.branch)
        wt.set_pending_merges(['b3'])
        wt.commit('merge b3', rev_id='a3')

        tree2.commit('unchanged', rev_id='b4')
        wt.branch.fetch(tree2.branch)
        wt.set_pending_merges(['b4'])

        self.assertEqual('a3', wt.last_revision())
        self.assertEqual(['b4'], wt.pending_merges())

        os.chdir('tree')
        out, err = self.run_bzr('uncommit', '--force', '-r', '2')

        self.assertEqual('a2', wt.last_revision())
        self.assertEqual(['b3', 'b4'], wt.pending_merges())

    def test_uncommit_octopus_merge(self):
        # Check that uncommit keeps the pending merges in the same order
        wt = self.create_simple_tree()

        tree2 = wt.bzrdir.sprout('tree2').open_workingtree()
        tree3 = wt.bzrdir.sprout('tree3').open_workingtree()

        tree2.commit('unchanged', rev_id='b3')
        tree3.commit('unchanged', rev_id='c3')
        wt.branch.fetch(tree2.branch)
        wt.branch.fetch(tree3.branch)
        wt.set_pending_merges(['b3', 'c3'])
        wt.commit('merge b3, c3', rev_id='a3')

        tree2.commit('unchanged', rev_id='b4')
        tree3.commit('unchanged', rev_id='c4')
        wt.branch.fetch(tree2.branch)
        wt.branch.fetch(tree3.branch)
        wt.set_pending_merges(['c4', 'b4'])
        wt.commit('merge b4, c4', rev_id='a4')

        self.assertEqual('a4', wt.last_revision())
        self.assertEqual([], wt.pending_merges())

        os.chdir('tree')
        out, err = self.run_bzr('uncommit', '--force', '-r', '2')

        self.assertEqual('a2', wt.last_revision())
        self.assertEqual(['b3', 'c3', 'c4', 'b4'], wt.pending_merges())
